from types import SimpleNamespace

import time

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.utils.data import Dataset, DataLoader

import torchvision
from torchvision import transforms
from torchvision import datasets
from torchvision.transforms import ToTensor
from torchvision.transforms.functional import InterpolationMode
from torchvision.utils import save_image

from activation import nn_size

import matplotlib.pyplot as plt

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

import pathlib

dataset_storage_path = '/home/users/r/rohrbla6/torch-dataset/'
project_root_dir = '/home/users/r/rohrbla6/activations/'
# dataset_storage_path = '/home/doek/torch-dataset/'
# project_root_dir = '/home/doek/activations/'

def path_drop_last(path):
	return "/".join(path.split("/")[:-1])

def mkdir(path):
	pathlib.Path(f'{project_root_dir}/{path}').mkdir(parents=True, exist_ok=True)

def savefig(path):
	mkdir(path_drop_last(path))
	plt.savefig(f"{project_root_dir}/{path}")

def saveimg(tensor,path):
	mkdir(path_drop_last(path))
	save_image(tensor,f"{project_root_dir}/{path}")

def savenp(np_array,path):
	mkdir(path_drop_last(path))
	np.save(f"{project_root_dir}/{path}",np_array)

# def save(model,path):
# 	mkdir(path_drop_last(path))
# 	torch.save(cnn, f"{project_root_dir}/{path}")

# def load(path):
# 	return torch.load(f"{project_root_dir}/{path}")

dataset = datasets.FashionMNIST(
	root = dataset_storage_path, download = True,
	train = True,
	transform=transforms.Compose([transforms.ToTensor(),]),
)

dataloader = DataLoader(dataset, batch_size=256, drop_last=False,)

dataset_t = datasets.FashionMNIST(
	root = dataset_storage_path, download = True,
	train = False,
	transform=transforms.Compose([transforms.ToTensor(),]),
)

dataloader_t = DataLoader(dataset_t, batch_size=256, drop_last=False,)

def autoenc_gen(σ=nn.ReLU(inplace=True), channels=128, latent_dim=16):
	enc = nn.Sequential(*[
		nn.Conv2d(1       ,channels  ,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.Conv2d(channels,channels  ,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.Conv2d(channels,channels  ,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.Conv2d(channels,channels  ,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.Conv2d(channels,channels  ,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.Conv2d(channels,channels  ,(3, 3),padding=0,stride=3),σ, nn.BatchNorm2d(channels),
		nn.Conv2d(channels,channels  ,(3, 3),padding=0,stride=3),σ, nn.BatchNorm2d(channels),
		nn.Conv2d(channels,latent_dim,(2, 2),padding=0,stride=1),σ, nn.BatchNorm2d(latent_dim),
	]).to(device)
	dec = nn.Sequential(*[
		nn.ConvTranspose2d(latent_dim,channels,(2, 2),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0,stride=3),σ, nn.BatchNorm2d(channels),
		nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0,stride=3),σ, nn.BatchNorm2d(channels),
		nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0,stride=1),σ, nn.BatchNorm2d(channels),
		nn.ConvTranspose2d(channels  ,1       ,(3, 3),padding=0,stride=1),nn.Sigmoid()#σ, #nn.Dropout2d(0.2), # nn.BatchNorm2d(1),
	]).to(device)
	return nn.Sequential(*[enc,dec]).to(device),enc,dec

# def autoenc_gen(σ=nn.ReLU(inplace=True), channels=32, latent_dim=32):
#   enc = nn.Sequential(*[
#     nn.Conv2d(1       ,channels  ,(3, 3),padding=1),σ, nn.MaxPool2d(2), #nn.Dropout2d(0.2), # nn.BatchNorm2d(4),
#     nn.Conv2d(channels,channels  ,(3, 3),padding=0),σ, nn.MaxPool2d(2), #nn.Dropout2d(0.2), # nn.BatchNorm2d(8),
#     nn.Conv2d(channels,channels  ,(3, 3),padding=0),σ, nn.MaxPool2d(2), #nn.Dropout2d(0.2), # nn.BatchNorm2d(12),
#     nn.Conv2d(channels,latent_dim,(2, 2),padding=0),σ,                  #nn.Dropout2d(0.2), # nn.BatchNorm2d(4),
#   ]).to(device)
#   dec = nn.Sequential(*[
#     nn.ConvTranspose2d(latent_dim,channels,(2, 2),padding=0),σ,                                         #nn.Dropout2d(0.2), # nn.BatchNorm2d(8),
#     nn.UpsamplingNearest2d(scale_factor=2), nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0),σ, #nn.Dropout2d(0.2), # nn.BatchNorm2d(4),
#     nn.UpsamplingNearest2d(scale_factor=2), nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0),σ, #nn.Dropout2d(0.2), # nn.BatchNorm2d(4),
#     nn.UpsamplingNearest2d(scale_factor=2), nn.ConvTranspose2d(channels  ,1       ,(3, 3),padding=1),σ, #nn.Dropout2d(0.2), # nn.BatchNorm2d(1),
#   ]).to(device)
#   return nn.Sequential(*[enc,dec]).to(device),enc,dec

# def autoenc_gen(σ=nn.ReLU(inplace=True), channels=32, latent_dim=32):
#   enc = nn.Sequential(*[
#     nn.Conv2d(1       ,channels  ,(3, 3),padding=1),σ, nn.MaxPool2d(2), nn.BatchNorm2d(channels),
#     nn.Conv2d(channels,channels  ,(3, 3),padding=0),σ, nn.MaxPool2d(2), nn.BatchNorm2d(channels),
#     nn.Conv2d(channels,channels  ,(3, 3),padding=0),σ, nn.MaxPool2d(2), nn.BatchNorm2d(channels),
#     nn.Conv2d(channels,latent_dim,(2, 2),padding=0),σ,                  nn.BatchNorm2d(channels),
#   ]).to(device)
#   dec = nn.Sequential(*[
#                                             nn.ConvTranspose2d(latent_dim,channels,(2, 2),padding=0),σ, nn.BatchNorm2d(channels),
#     nn.UpsamplingNearest2d(scale_factor=2), nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0),σ, nn.BatchNorm2d(channels),
#     nn.UpsamplingNearest2d(scale_factor=2), nn.ConvTranspose2d(channels  ,channels,(3, 3),padding=0),σ, nn.BatchNorm2d(channels),
#     nn.UpsamplingNearest2d(scale_factor=2), nn.ConvTranspose2d(channels  ,1       ,(3, 3),padding=1),nn.Sigmoid()
#   ]).to(device)
#   return nn.Sequential(*[enc,dec]).to(device),enc,dec

def debug_autoenc_gen():
	autoenc,enc,dec = autoenc_gen()
	for i_batch, sample_batched in enumerate(dataloader):
		x = sample_batched[0].to(device)
		# x = dataset[0][0].to(device)
		print(nn_size(autoenc))
		print(x.shape)
		print(enc(x).shape)
		print(autoenc(x).shape)
		loss = F.mse_loss(autoenc(x),x)
		print(loss)
		loss.backward()
		break

debug_autoenc_gen()

autoenc_loss_conv = nn.Conv2d(1,5,(3, 3))
autoenc_loss_pad = nn.ReflectionPad2d(1)

with torch.no_grad():
	λ = 4
	autoenc_loss_conv.weight[0] = torch.tensor([[0,0,0],[0,λ,0],[0,0,0]])         # weighted identity (weighted so that is important on the loss)
	autoenc_loss_conv.weight[1] = torch.tensor([[1,2,1],[0,0,0],[-1,-2,-1]])  /4. # derivative
	autoenc_loss_conv.weight[2] = torch.tensor([[1,2,1],[0,0,0],[-1,-2,-1]]).T/4. # derivative other direction
	autoenc_loss_conv.weight[3] = torch.tensor([[1,1,1],[1,1,1],[ 1, 1, 1]]).T/9. # box blur
	autoenc_loss_conv.weight[4] = torch.tensor([[1,2,1],[2,4,2],[ 1, 2, 1]]).T/16.# gaussian blur

autoenc_loss_conv = autoenc_loss_conv.to(device)

def autoenc_loss(neuralnet,data):
	per_pixel_and_grad = F.mse_loss(
		autoenc_loss_conv(autoenc_loss_pad(neuralnet(data))),
		autoenc_loss_conv(autoenc_loss_pad(data))
	)
	return per_pixel_and_grad


# for sample_batched in dataloader:
#   autoenc_loss(autoenc,sample_batched[0].to(device))
#   break

def autoenc_setup(neural_network,lr=1e-4):
	o = SimpleNamespace() # output
	o.nn = neural_network
	o.optimizer = torch.optim.Adam(o.nn.parameters(), lr)
	o.hist_loss = []
	o.hist_loss_b = []
	return o

def train(nn_obj,epochs,name,snapshot):
	# with alive_bar(epochs,force_tty=True) as bar:
		t0 = time.time()
		for epoch in range(epochs):
			hist_loss=0
			hist_loss_b=0
			count=0
			for i_batch, sample_batched in enumerate(dataloader):
				x = sample_batched[0].to(device)
				nn_obj.nn.train()
				nn_obj.optimizer.zero_grad()
				# loss = F.mse_loss(nn_obj.nn(x),x)
				loss = autoenc_loss(nn_obj.nn,x)
				loss.backward()
				nn_obj.optimizer.step()
			for i_batch, sample_batched in enumerate(dataloader_t):
				count+=1
				with torch.no_grad():
					nn_obj.nn.eval()
					loss = autoenc_loss(nn_obj.nn,x)
					loss_b = F.mse_loss(nn_obj.nn(x), x)
					hist_loss += loss.item()
					hist_loss_b += loss_b.item()
			hist_loss/=count
			hist_loss_b/=count
			nn_obj.hist_loss.append(hist_loss)
			nn_obj.hist_loss_b.append(hist_loss_b)
			# print(f"progress : {epoch}/{epochs} = {epoch/epochs}", end="\r")
			print(f"progress : {epoch}/{epochs} = {epoch/epochs} | duration={time.time()-t0}")
			if (epoch<=16) or (epoch==epochs-1) or ((epoch % snapshot) == 0) : viz(nn_obj,f"loss {name}",f"autoenc-cnn/{name}/epoch={epoch}.png")
			viz_loss(nn_obj,f"loss {name}",f"autoenc-cnn/svg/{name}-epochs={epochs}.svg")
			viz_loss(nn_obj,f"loss {name}",f"autoenc-cnn/{name}-epochs={epochs}.png")
			if (epoch%16 == 0): torch.cuda.empty_cache()
			savenp(np.array(nn_obj.hist_loss),f"loss/{name}-diverse-kernels.np")
			savenp(np.array(nn_obj.hist_loss_b),f"loss/{name}-per-pixel-only.np")
		# print("")
	# 		bar()

def imshow(image_tensor,nrow=16):
	grid_img = torchvision.utils.make_grid(image_tensor.to('cpu'), nrow=nrow)
	ratio = grid_img.shape[2]/grid_img.shape[1]
	plt.figure(figsize=(30,30*ratio))
	plt.imshow(grid_img.permute(1, 2, 0))

def viz(nn_obj,title="",path=None):
	for i_batch, sample_batched in enumerate(dataloader):
		batch_tensor = nn_obj.nn(sample_batched[0].to(device)).to('cpu')
		#imshow(batch_tensor)
		#if path!=None: savefig(f"images-{path}")
		if path!=None:
			saveimg(torchvision.utils.make_grid(batch_tensor.to('cpu'), nrow=16),f"images-trainset-{path}")
			saveimg(torchvision.utils.make_grid(torch.abs(batch_tensor-sample_batched[0]).to('cpu'), nrow=16),f"error-trainset-{path}")
		else:
			imshow(batch_tensor)
		break
	for i_batch, sample_batched in enumerate(dataloader_t):
		batch_tensor = nn_obj.nn(sample_batched[0].to(device)).to('cpu')
		if path!=None:
			saveimg(torchvision.utils.make_grid(batch_tensor.to('cpu'), nrow=16),f"images-{path}")
			saveimg(torchvision.utils.make_grid(torch.abs(batch_tensor-sample_batched[0]).to('cpu'), nrow=16),f"error-{path}")
		else:
			imshow(batch_tensor)
		break
	plt.show()

def viz_loss(nn_obj,title="",path=None):
	plt.figure(figsize=(10,5))
	# plt.clf()
	plt.plot(np.arange(len(nn_obj.hist_loss))+1,np.log(nn_obj.hist_loss),label="log(loss)")
	# plt.plot(nn_obj.hist_loss_b,label="loss per pixel")
	plt.legend(loc="upper right")
	plt.grid()
	# plt.yscale('log') # did put log in the evaluation so that axis looks better
	plt.xscale('log')
	plt.title(title)
	if path!=None: savefig(path)

def σviz(σ,name,path=None,clf=True,color=[0,0,1]):
	x = torch.linspace(-2,2,2000).to(device)
	if clf: plt.clf()
	plt.plot(
		x.to('cpu').detach().numpy(),
		σ(x).to('cpu').detach().numpy(),
		linewidth=.5,color=color
	)
	plt.title(name)
	if path==None and clf==True: plt.show()
	else:
		savefig(f"σ/{path}.png")
		savefig(f"σ/{path}.svg")

def benchmark(name,σ,epochs):
	print(f"autoencoder {name} ...")
	cnn = autoenc_setup(autoenc_gen(σ)[0])
	train(cnn,epochs,name,1)
	# save(cnn, f"autoenc-cnn-model/{name}.torch")
	# viz(cnn,f"loss {name}",f"autoenc-cnn/{name}-epochs={epochs}.png")
	# viz(cnn,f"loss {name}",f"autoenc-cnn/svg/{name}-epochs={epochs}.svg")
	plt.figure(figsize=(10,5))
	σviz(σ,name,name)
	print(f"done")
	# savenp(np.array(cnn.hist_loss),f"loss/{name}-diverse-kernels.np")
	# savenp(np.array(cnn.hist_loss_b),f"loss/{name}-per-pixel-only.np")
	# print(f"save done")
	plt.close('all')

